#include <Wire.h>
#include <LiquidCrystal_I2C.h>
#include "WiFi.h"
#include <Preferences.h>

LiquidCrystal_I2C lcd(0x27, 20, 4);
String ssid     = "";
String password = "";
String CallSign   = "";

String host = "";
int port = 23;
const int timeout = 50000;

#define console Serial
#define consoleSpeed 115200
String inputCommands = "";
bool   inputCommandsComplete = false;
const byte inputCommandsLength = 60;
unsigned long serialEventTime = 0;

int value = 0;
WiFiClient client;
Preferences prefs;

char* string2char(String command) {
  if (command.length() != 0) {
    char *p = const_cast<char*>(command.c_str());
    return p;
  }
}

void setup()
{
  console.begin(consoleSpeed);
  prefs.begin("nvs", false);
  ssid =  prefs.getString("ssid_name");
  password = prefs.getString("ssid_pass");
  CallSign = prefs.getString("callsign");
  host = prefs.getString("host");
  port = prefs.getInt("port");

  lcd.init();
  lcd.backlight();
  connectToWifi();
}

void connectToWifi()
{
  uint64_t start = millis();
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("Connecting to ");
  lcd.setCursor(0, 1);
  lcd.print(ssid);

  WiFi.begin(string2char(ssid), string2char(password));

  while (WiFi.status() != WL_CONNECTED) {
    delay(500);
    lcd.print(".");
    if (millis() - start > timeout)
    {
      WiFi.disconnect();
      lcd.clear();
      lcd.print("WiFi not connected!");
      break;
    }
  }
  if (WiFi.status() == WL_CONNECTED)
  {
    lcd.clear();
    lcd.setCursor(0, 0);
    lcd.print("WiFi connected");
    lcd.setCursor(0, 1);
    lcd.print("IP address: ");
    lcd.setCursor(0, 2);
    lcd.print(WiFi.localIP());
    connectToDX();
  }
}

void connectToDX()
{
  if (!client.connect(string2char(host), port)) {
    lcd.setCursor(0, 3);
    lcd.print("Disconnect cluster");
    return;
  } else {
    lcd.setCursor(0, 3);
    lcd.print("Connect cluster OK");
    delay(10);
    client.print(CallSign + "\n");
    delay(1000);
    lcd.clear();
  }
}

void WifiScan()
{
  WiFi.mode(WIFI_STA);
  WiFi.disconnect();
  int n = WiFi.scanNetworks();
  if (n == 0) {
    console.println("no networks found");
  } else {
    console.print("NETWORKS:");
    console.print(n);
    console.print(",");
    for (int i = 0; i < n; ++i) {
      console.print(WiFi.SSID(i));
      console.print(",");
      delay(10);
    }
  }
  Serial.println("");
}

void serialEvent() {
  serialEventTime = millis();
  if (console.available()) {
    char c = (char)console.read();
    if (inputCommands.length() < inputCommandsLength) {
      if (c != '\n') inputCommands += c;
      else if (inputCommands.length()) inputCommandsComplete = true;
    }
  }
}

void parseBytes(const char* str, char sep, byte* bytes, int maxBytes, int base) {
  for (int i = 0; i < maxBytes; i++) {
    bytes[i] = strtoul(str, NULL, base);
    str = strchr(str, sep);
    if (str == NULL || *str == '\0') break;
    str++;
  }
}

void serialEventControl() {
  if (serialEventTime) {
    if (millis() - serialEventTime > 1000 or serialEventTime > millis()) {
      inputCommands = "";
      serialEventTime = 0;
    }
  }
}
void waitAndPrint(String text, uint8_t time) {
  console.println(text);
  delay(time);
}

void resetFunc()
{
  ESP.restart();
}

void clearParam()
{
  prefs.putString("ssid_name", "");
  prefs.putString("ssid_pass", "");
  prefs.putString("callsign", "");
}

void serialEventHandler() {
  if (inputCommandsComplete) {
    inputCommands.trim();
    console.println(inputCommands);
    if (inputCommands == F("HELP")) {
      consoleHelp();
    } else if (inputCommands == F("SCAN")) {
      WifiScan();

    } else if (inputCommands == F("CONNECT_WIFI")) {
      connectToWifi();

    } else if (inputCommands == F("CONNECT_DX")) {
      connectToDX();

    } else if (inputCommands.startsWith(F("SET_SSID"))) {
      String new_ssid = inputCommands.substring(8);
      new_ssid.trim();
      if (new_ssid.length() < 40) {
        prefs.putString("ssid_name", new_ssid);
        ssid =  prefs.getString("ssid_name");
        console.println(ssid);
        console.println(F("# ok"));
      } else console.println(F("# length must not exceed 40 characters"));


    } else if (inputCommands.startsWith(F("SET_PASS"))) {
      String new_pass = inputCommands.substring(8);
      new_pass.trim();
      if (new_pass.length() < 40) {
        prefs.putString("ssid_pass", new_pass);
        password =  prefs.getString("ssid_pass");
        console.println(password);
        console.println(F("# ok"));
      } else console.println(F("# length must not exceed 40 characters"));

    } else if (inputCommands.startsWith(F("SET_CALL"))) {
      String new_call = inputCommands.substring(8);
      new_call.trim();
      if (new_call.length() < 40) {
        prefs.putString("callsign", new_call);
        CallSign =  prefs.getString("callsign");
        console.println(CallSign);
        console.println(F("# ok"));
      } else console.println(F("# length must not exceed 40 characters"));

    } else if (inputCommands.startsWith(F("SET_HOST"))) {
      String new_host = inputCommands.substring(8);
      new_host.trim();
      prefs.putString("host", new_host);
      host =  prefs.getString("host");
      console.println(host);
      console.println(F("# ok"));

    } else if (inputCommands.startsWith(F("SET_PORT"))) {
      String new_port = inputCommands.substring(8);
      new_port.trim();
      prefs.putInt("port", new_port.toInt());
      port =  prefs.getInt("port", 23);
      console.println(port);
      console.println(F("# ok"));

    } else if (inputCommands == F("RESTART")) {
      waitAndPrint(F("# ok\n# restarting device..."), 100);
      resetFunc();
    } else if (inputCommands == F("??")) {
      printConfig();
    } else if (inputCommands == F("CLEAR_PARAM")) {
      console.println(F("# eraseing..."));
      clearParam();
      console.println(F("# done"));

    }  else console.println(F("# bad command"));
    inputCommands = "";
    inputCommandsComplete = false;
    serialEventTime = 0;
  }
  serialEventControl();
}

void loop()
{
  serialEvent();
  serialEventHandler();
  while (client.available()) {
    PrintDX(client.readStringUntil('\r'));
  }
}

void PrintDX(String TelnetLine)
{
  Serial.print(TelnetLine);
  TelnetLine.trim();
  if (TelnetLine.indexOf("DX de") >= 0)
  {
    lcd.clear();
    TelnetLine.replace(":", " ");
    lcd.setCursor(0, 0);
    lcd.print("DX de ");
    lcd.print(TelnetLine.substring(6, 14));
    lcd.setCursor(0, 1);
    lcd.print("Call  ");
    lcd.print(TelnetLine.substring(26, 38));
    lcd.setCursor(0, 2);
    lcd.print("Freq ");
    lcd.print(TelnetLine.substring(15, 24));
    lcd.setCursor(0, 3);
    lcd.print("Time  ");
    lcd.print(TelnetLine.substring(70, 72));
    lcd.print(":");
    lcd.print(TelnetLine.substring(72, 74));
  }
}

void printConfig()
{
  console.print("CONFIG");
  console.print(',');
  console.print(ssid);
  console.print(',');
  console.print(password);
  console.print(',');
  console.print(CallSign);
  console.print(',');
  console.print(host);
  console.print(',');
  console.print(port);
  console.print(',');
  console.print(WiFi.localIP());
  console.print(',');
  console.print(WiFi.subnetMask());
  console.print(',');
  console.print(WiFi.gatewayIP());
  console.print(',');
  Serial.println("");
}

void consoleHelp() {
  console.println(F("- ---------------------------------------------------------------------------------------"));
  console.println(F("# Additional commands:"));
  console.println(F("# SET_CALL - Set Callsign"));
  console.println(F("# SET_HOST - Set host DX cluster"));
  console.println(F("# SET_PORT - Set port DX Cluster"));
  console.println(F("# SET_SSID - Set WiFi network name"));
  console.println(F("# SET_PASS - Set WiFi SSID password"));
  console.println(F("# RESTART - restarts the device"));
  console.println(F("# CONNECT_WIFI - Connect to WiFi network"));
  console.println(F("# CONNECT_DX - Connect to DX Cluster"));
  console.println(F("# SCAN - Scan WiFi networks"));
  console.println(F("# CLEAR_PARAM - Clear all saving parameters"));
  console.println(F("# HELP - view this help"));
  console.println(F("- ---------------------------------------------------------------------------------------"));
}
